# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import pytest

from pikepdf import Dictionary, Name, Object, Pdf, Stream

# pylint: disable=e1137


def test_create_form_xobjects(outdir):
    pdf = Pdf.new()

    font = pdf.make_indirect(
        Object.parse(
            b"""
            <<
                /Type /Font
                /Subtype /Type1
                /Name /F1
                /BaseFont /Helvetica
                /Encoding /WinAnsiEncoding
            >>
        """
        )
    )

    width, height = 100, 100
    image_data = b"\xff\x7f\x00" * (width * height)

    image = Stream(pdf, image_data)
    image.stream_dict = Object.parse(
        """
            <<
                /Type /XObject
                /Subtype /Image
                /ColorSpace /DeviceRGB
                /BitsPerComponent 8
                /Width 100
                /Height 100
            >>
    """
    )
    xobj_image = Dictionary({'/Im1': image})

    form_xobj_res = Dictionary({'/XObject': xobj_image})
    form_xobj = Stream(
        pdf,
        b"""
        /Im1 Do
    """,
    )
    form_xobj['/Type'] = Name('/XObject')
    form_xobj['/Subtype'] = Name('/Form')
    form_xobj['/FormType'] = 1
    form_xobj['/Matrix'] = [1, 0, 0, 1, 0, 0]
    form_xobj['/BBox'] = [0, 0, 1, 1]
    form_xobj['/Resources'] = form_xobj_res

    rfont = {'/F1': font}

    resources = {'/Font': rfont, '/XObject': {'/Form1': form_xobj}}

    mediabox = [0, 0, 612, 792]

    stream = b"""
        BT /F1 24 Tf 72 720 Td (Hi there) Tj ET
        q 144 0 0 144 234 324 cm /Form1 Do Q
        q 72 0 0 72 378 180 cm /Form1 Do Q
        """

    contents = Stream(pdf, stream)

    page = pdf.make_indirect(
        {
            '/Type': Name('/Page'),
            '/MediaBox': mediabox,
            '/Contents': contents,
            '/Resources': resources,
        }
    )

    pdf.pages.append(page)
    pdf.save(outdir / 'formxobj.pdf')



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import datetime
import os
from hashlib import md5
from pathlib import Path

import pytest

import pikepdf
from pikepdf import Dictionary, Name, Pdf
from pikepdf._qpdf import AttachedFile, AttachedFileSpec, Attachments


@pytest.fixture
def pal(resources):
    yield Pdf.open(resources / 'pal.pdf')


def test_attachment_crud(pal, resources, outpdf):
    assert hasattr(pal, 'attachments'), "no attachments interface"
    assert len(pal.attachments) == 0, "test file already has attachments"
    assert 'anything' not in pal.attachments, "attachment interface is quirky"

    rle = resources / 'rle.pdf'
    fs = AttachedFileSpec.from_filepath(pal, rle)
    rle_bytes = rle.read_bytes()

    pal.attachments['rle.pdf'] = fs

    assert len(pal.attachments) == 1, "attachment count not incremented"
    assert 'rle.pdf' in pal.attachments, "attachment filename not registered"
    assert 'attached' in repr(pal.attachments)

    pal.save(outpdf)

    with Pdf.open(outpdf) as output:
        assert len(output.attachments) == 1, "output had no attachment"
        assert 'rle.pdf' in output.attachments, "filename not present"
        rle_spec = output.attachments['rle.pdf']
        rle_file = rle_spec.get_file()
        assert (
            rle_bytes == rle_file.read_bytes()
        ), "attachment not reproduced bit for bit"

        del output.attachments['rle.pdf']
        assert 'rle.pdf' not in output.attachments, "del failed"
        assert len(output.attachments) == 0, "not removed"

    with Pdf.open(outpdf) as output:
        output.attachments.pop('rle.pdf')


def test_attachment_iter(pal):
    inputs = ['1', '2']

    for input_ in inputs:
        pal.attachments[f'filename {input_}'] = AttachedFileSpec(
            pal, input_.encode('ascii')
        )

    for filename in pal.attachments:
        fileno = filename.replace('filename ', '')
        assert fileno in inputs
        filespec = pal.attachments[filename]
        assert filespec.get_file().read_bytes().decode('ascii') in inputs


def test_filespec_types(pal, resources):
    some_bytes = b'just some bytes'
    some_path = resources / 'rle.pdf'
    assert isinstance(some_path, Path)

    fs_bytes = AttachedFileSpec(pal, some_bytes)
    assert fs_bytes.get_file().read_bytes() == some_bytes
    assert fs_bytes.filename == ''

    fs_path = AttachedFileSpec.from_filepath(pal, some_path)
    assert fs_path.get_file().read_bytes() == some_path.read_bytes()

    with pytest.raises(TypeError):
        fs_path.get_file(pikepdf.Array([1]))


def test_attachment_metadata(pal, data=b'some data', description='test filespec'):
    fs = AttachedFileSpec(pal, data, description=description)

    assert fs.description == description
    fs.description = 'xyz'
    assert fs.description == 'xyz'

    attached_stream = fs.get_file()
    assert attached_stream.creation_date is None
    assert attached_stream.mod_date is None

    june_1 = datetime.datetime(2021, 6, 1, 1, 2, 3)

    attached_stream.creation_date = june_1
    attached_stream.mod_date = june_1
    assert attached_stream.creation_date == june_1
    assert attached_stream.mod_date == june_1

    assert attached_stream.size == len(data)

    assert attached_stream.mime_type == ''
    attached_stream.mime_type = 'text/plain'
    assert attached_stream.mime_type == 'text/plain'


def test_compound_attachment(pal):
    data = [b'data stream 1', b'data stream 2']
    filename = [b'filename of data stream 1', b'filename of data stream 2']

    fs = AttachedFileSpec(pal, data[0], description='test filespec')
    fs.filename = 'compound filespec'
    fs.obj.UF = filename[0]

    # Add another data stream to the underlying object
    fs.obj.EF.F = pal.make_stream(data[1], Type=Name.EmbeddedFile)
    fs.obj.F = filename[1]

    all_filenames = fs.get_all_filenames()
    assert len(all_filenames) == 2
    assert Name.F in all_filenames
    assert Name.UF in all_filenames
    assert all_filenames[Name.UF] == filename[0]
    assert all_filenames[Name.F] == filename[1]

    assert fs.get_file().read_bytes() == data[0]
    assert fs.get_file(Name.UF).read_bytes() == data[0]
    assert fs.get_file(Name.F).read_bytes() == data[1]

    assert fs.get_file().md5 == md5(data[0]).digest()

    pal.attachments['compound'] = fs


def test_from_str_filepath(pal, outdir):
    foofile = outdir / 'foo'
    foofile.touch()
    fs = AttachedFileSpec.from_filepath(pal, str(foofile), description='bar')
    assert os.sep not in fs.filename
    assert 'foo' in repr(fs)
    fs.filename = ''
    assert 'foo' not in repr(fs)
    assert 'AttachedFile' in repr(fs.get_file())



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

"""
Testing focused on pikepdf.Pdf
"""

from __future__ import annotations

import locale
import shutil
import zlib
from contextlib import nullcontext
from io import BytesIO, StringIO
from os import fspath
from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import Mock

import pytest
from conftest import fails_if_pypy

import pikepdf
from pikepdf import Name, PasswordError, Pdf, PdfError, Stream
from pikepdf._exceptions import DependencyError

# pylint: disable=redefined-outer-name


@pytest.fixture
def trivial(resources):
    with Pdf.open(
        resources / 'pal-1bit-trivial.pdf'  # , access_mode=pikepdf.AccessMode.mmap
    ) as pdf:
        yield pdf


@pytest.fixture
def pdf_form(resources):
    with Pdf.open(resources / 'form.pdf') as pdf:
        yield pdf


def test_new(outdir):
    pdf = pikepdf.new()
    pdf.save(outdir / 'new-empty.pdf')


def test_non_filename():
    with pytest.raises(TypeError):
        Pdf.open(42.0)


def test_file_descriptor(resources):
    with (resources / 'pal-1bit-trivial.pdf').open('rb') as f:
        with pytest.raises(TypeError):
            Pdf.open(f.fileno())


def test_save_to_file_descriptor_fails(trivial):
    with pytest.raises(TypeError):
        trivial.save(2)


def test_not_existing_file():
    with pytest.raises(FileNotFoundError):
        Pdf.open('does_not_exist.pdf')


def test_empty(outdir):
    target = outdir / 'empty.pdf'
    target.touch()
    with pytest.raises(PdfError):
        Pdf.open(target)


class TestLinearization:
    def test_linearization(self, resources, outdir):
        with Pdf.open(resources / 'graph.pdf') as graph:
            assert not graph.is_linearized
            graph.save(outdir / 'lin.pdf', linearize=True)

        with Pdf.open(outdir / 'lin.pdf') as pdf:
            assert pdf.is_linearized

            sio = StringIO()
            pdf.check_linearization(sio)


def test_objgen(resources):
    with Pdf.open(resources / 'graph.pdf') as src:
        im0 = src.pages[0].Resources.XObject['/Im0']
        assert im0.objgen == (5, 0)
        object5 = src.get_object((5, 0))
        assert object5.is_owned_by(src)
        assert object5 == im0
        assert object5 == src.get_object(5, 0)


class TestPasswords:
    def test_open_pdf_wrong_password(self, resources):
        # The correct passwords are "owner" and "user"
        with pytest.raises(PasswordError):
            Pdf.open(resources / 'graph-encrypted.pdf', password='wrong')

    def test_open_pdf_password_encoding(self, resources):
        with pytest.raises(PasswordError):
            Pdf.open(resources / 'graph-encrypted.pdf', password=b'\x01\xfe')

    def test_open_pdf_no_password_but_needed(self, resources):
        with pytest.raises(PasswordError):
            Pdf.open(resources / 'graph-encrypted.pdf')

    def test_open_pdf_user_password(self, resources):
        with Pdf.open(resources / 'graph-encrypted.pdf', password='user'):
            pass

    def test_unneeded_password_ignored(self, resources):
        with pytest.warns(UserWarning, match="no password was needed"):
            with Pdf.open(resources / 'graph.pdf', password='open sesame'):
                pass


class TestPermissions:
    def test_some_permissions_missing(self, resources):
        with Pdf.open(resources / 'graph-encrypted.pdf', password='owner') as pdf:
            assert not pdf.allow.print_highres
            assert not pdf.allow.modify_annotation
            assert pdf.allow.print_lowres

    def test_all_true_not_encrypted(self, trivial):
        assert all(trivial.allow)

    def test_omit_encryption_removes_encryption(self, resources, outdir):
        with Pdf.open(resources / 'graph-encrypted.pdf', password='owner') as pdf:
            pdf.save(outdir / 'true.pdf')
            with Pdf.open(outdir / 'true.pdf') as pdf_copy:
                assert pdf.allow != pdf_copy.allow

    @pytest.mark.parametrize('encryption, expected', [[True, True], [False, False]])
    @pytest.mark.filterwarnings('ignore:A password was provided')
    def test_permissions_preserved_on_save(
        self, resources, outdir, encryption, expected
    ):
        with Pdf.open(resources / 'graph-encrypted.pdf', password='owner') as pdf:
            pdf.save(outdir / 'true.pdf', encryption=encryption)
            with Pdf.open(outdir / 'true.pdf', password='owner') as pdf_copy:
                assert (pdf.allow == pdf_copy.allow) == expected


class TestStreams:
    def test_stream(self, resources):
        with (resources / 'pal-1bit-trivial.pdf').open('rb') as stream:
            with Pdf.open(stream) as pdf:
                assert pdf.Root.Pages.Count == 1

    def test_no_text_stream(self, resources):
        with pytest.raises(TypeError):
            with (resources / 'pal-1bit-trivial.pdf').open('r') as stream:
                Pdf.open(stream)

    def test_save_stream(self, trivial, outdir):
        pdf = trivial
        pdf.save(outdir / 'nostream.pdf', deterministic_id=True)

        bio = BytesIO()
        pdf.save(bio, deterministic_id=True)
        bio.seek(0)

        with (outdir / 'nostream.pdf').open('rb') as saved_file:
            saved_file_contents = saved_file.read()
        assert saved_file_contents == bio.read()

    def test_read_not_readable_file(self, outdir):
        with (Path(outdir) / 'writeme.pdf').open('wb') as writable:
            with pytest.raises(ValueError, match=r'not readable'):
                Pdf.open(writable)

    def test_open_not_seekable_stream(self, resources):
        class UnseekableBytesIO(BytesIO):
            def seekable(self):
                return False

        testio = UnseekableBytesIO((resources / 'pal-1bit-trivial.pdf').read_bytes())

        with pytest.raises(ValueError, match=r'not seekable'):
            Pdf.open(testio)


class TestMemory:
    def test_memory(self, resources):
        data = (resources / 'pal-1bit-trivial.pdf').read_bytes()
        with pytest.warns(UserWarning, match="bytes-like object containing a PDF"):
            with pytest.raises(Exception):
                Pdf.open(data)


def test_remove_unreferenced(resources, outdir):
    in_ = resources / 'sandwich.pdf'
    out1 = outdir / 'out1.pdf'
    out2 = outdir / 'out2.pdf'
    with Pdf.open(in_) as pdf:
        pdf.pages[0].Contents = Stream(pdf, b' ')
        pdf.save(out1)

        pdf.remove_unreferenced_resources()
        pdf.save(out2)

    assert out2.stat().st_size < out1.stat().st_size


def test_show_xref(trivial):
    trivial.show_xref_table()


def test_progress(trivial, outdir):
    pdf = trivial
    mock = Mock()
    pdf.save(outdir / 'out.pdf', progress=mock)
    mock.assert_called()


@pytest.mark.skipif(locale.getpreferredencoding() != 'UTF-8', reason="Unicode check")
def test_unicode_filename(resources, outdir):
    target1 = outdir / '测试.pdf'  # Chinese: test.pdf
    target2 = outdir / '通过考试.pdf'  # Chinese: pass the test.pdf
    shutil.copy(fspath(resources / 'pal-1bit-trivial.pdf'), fspath(target1))
    with Pdf.open(target1) as pdf:
        pdf.save(target2)
    assert target2.exists()


def test_min_and_force_version(trivial, outdir):
    pdf = trivial
    pdf.save(outdir / '1.7.pdf', min_version='1.7')

    with Pdf.open(outdir / '1.7.pdf') as pdf17:
        assert pdf17.pdf_version == '1.7'

    pdf.save(outdir / '1.2.pdf', force_version='1.2')

    with Pdf.open(outdir / '1.2.pdf') as pdf12:
        assert pdf12.pdf_version == '1.2'


def test_normalize_linearize(trivial, outdir):
    with pytest.raises(ValueError):
        trivial.save(outdir / 'no.pdf', linearize=True, normalize_content=True)


def test_make_stream(trivial, outdir):
    pdf = trivial
    stream = pdf.make_stream(b'q Q')
    pdf.pages[0].Contents = stream
    pdf.save(outdir / 's.pdf')


def test_add_blank_page(trivial):
    assert len(trivial.pages) == 1

    invalid = [-1, 0, 2, 15000]
    for n in invalid:
        with pytest.raises(ValueError):
            trivial.add_blank_page(page_size=(n, n))
    trivial.add_blank_page()
    assert len(trivial.pages) == 2


def test_object_stream_mode_generated(trivial, outdir):
    trivial.save(
        outdir / '1.pdf',
        fix_metadata_version=True,
        object_stream_mode=pikepdf.ObjectStreamMode.generate,
    )
    assert b'/ObjStm' in (outdir / '1.pdf').read_bytes()

    trivial.save(
        outdir / '2.pdf',
        fix_metadata_version=False,
        object_stream_mode=pikepdf.ObjectStreamMode.generate,
    )
    assert b'/ObjStm' in (outdir / '2.pdf').read_bytes()


def test_with_block(resources):
    desc = ''
    with pikepdf.open(resources / 'pal-1bit-trivial.pdf') as pdf:
        desc = pdf.filename
    assert pdf.filename != desc


def test_closed_anon_pdf():
    pdf = pikepdf.new()
    desc = pdf.filename
    pdf.close()
    assert pdf.filename != desc


def test_with_block_abuse(resources):
    with pikepdf.open(resources / 'pal-1bit-trivial.pdf') as pdf:
        im0 = pdf.pages[0].Resources.XObject['/Im0']
    with pytest.raises(PdfError):
        im0.read_bytes()


def test_allow_overwriting_input(resources, tmp_path):
    orig_pdf_path = fspath(resources / 'pal-1bit-trivial.pdf')
    tmp_pdf_path = fspath(tmp_path / 'pal-1bit-trivial.pdf')
    shutil.copy(orig_pdf_path, tmp_pdf_path)
    with pikepdf.open(tmp_pdf_path, allow_overwriting_input=True) as pdf:
        with pdf.open_metadata() as meta:
            meta['dc:title'] = 'New Title'
        pdf.save('other.pdf', encryption=dict(owner="owner"))
        pdf.save()
        pdf.save(linearize=True)
    with pikepdf.open(tmp_pdf_path) as pdf:
        with pdf.open_metadata() as meta:
            assert meta['dc:title'] == 'New Title'
    with pikepdf.open(orig_pdf_path) as pdf:
        with pdf.open_metadata() as meta:
            assert 'dc:title' not in meta


def test_allow_overwriting_input_without_filename():
    with pytest.raises(ValueError):
        with pikepdf.open(BytesIO(), allow_overwriting_input=True):
            pass


def test_allow_overwriting_input_from_pdf_new():
    pdf = Pdf.new()
    with pytest.raises(ValueError, match="allow_overwriting_input=True"):
        pdf.save()


def test_check(resources):
    with pikepdf.open(resources / 'content-stream-errors.pdf') as pdf:
        problems = pdf.check()
        assert len(problems) > 0
        assert all(isinstance(prob, str) for prob in problems)
        assert 'parse error while reading' in problems[0]


@fails_if_pypy
def test_check_specialized_decoder_fallback(resources, monkeypatch):
    class MissingJBIG2Decoder(pikepdf.jbig2.JBIG2DecoderInterface):
        def check_available(self):
            raise DependencyError('jbig2dec')

        def decode_jbig2(self, jbig2: bytes, jbig2_globals: bytes) -> bytes:
            raise CalledProcessError(1, 'jbig2dec')

    monkeypatch.setattr(pikepdf.jbig2, 'get_decoder', MissingJBIG2Decoder)

    with pikepdf.open(resources / 'jbig2.pdf') as pdf:
        with pytest.warns(UserWarning, match=r".*missing some specialized.*"):
            problems = pdf.check()
        assert len(problems) == 0


def test_repr(trivial):
    assert repr(trivial).startswith('<')


def test_recompress(resources, outdir):
    with pikepdf.open(resources / 'image-mono-inline.pdf') as pdf:
        obj = pdf.get_object((7, 0))
        assert isinstance(obj, pikepdf.Stream)

        data = obj.read_bytes()
        data_z1 = zlib.compress(data, level=0)  # No compression but zlib wrapper

        obj.write(data_z1, filter=pikepdf.Name.FlateDecode)

        bigger = outdir / 'a.pdf'
        smaller = outdir / 'b.pdf'
        pdf.save(bigger, recompress_flate=False)
        pdf.save(smaller, recompress_flate=True)
        assert smaller.stat().st_size < bigger.stat().st_size


def test_invalid_flate_compression_level():
    # We don't want to change the compression level because it's global state
    # and will change subsequent test results, so just ping it with an invalid
    # value to get partial code coverage.
    with pytest.raises(ValueError):
        pikepdf.settings.set_flate_compression_level(99)


def test_flate_compression_level():
    # While this function affects global state, we can test it safely because
    # setting the value to -1 restores the default.
    try:
        pikepdf.settings.set_flate_compression_level(0)
        pikepdf.settings.set_flate_compression_level(9)
    finally:
        pikepdf.settings.set_flate_compression_level(-1)


def test_set_access_default_mmap():
    initial = pikepdf._qpdf.get_access_default_mmap()
    try:
        pikepdf._qpdf.set_access_default_mmap(True)
    finally:
        pikepdf._qpdf.set_access_default_mmap(initial)


def test_generate_appearance_streams(pdf_form):
    assert Name.AP not in pdf_form.Root.AcroForm.Fields[0]

    pdf_form.Root.AcroForm.NeedAppearances = True
    pdf_form.generate_appearance_streams()

    assert Name.AP in pdf_form.Root.AcroForm.Fields[0]


@pytest.mark.parametrize(
    'mode, exc',
    [('all', None), ('print', None), ('screen', None), ('', None), ('42', ValueError)],
)
def test_flatten_annotations_parameters(pdf_form, mode, exc):

    if exc is not None:
        error_ctx = pytest.raises(exc)
    else:
        error_ctx = nullcontext()
    with error_ctx:
        if mode is None:
            pdf_form.flatten_annotations()
        else:
            pdf_form.flatten_annotations(mode)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

with open('pyproject.toml', 'rb') as f:
    t = tomllib.load(f)
env = t['tool']['cibuildwheel']['environment']
print('\n'.join(f'{k}={v}' for k, v in env.items()))



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import pytest

from pikepdf import Pdf

# pylint: disable=redefined-outer-name,pointless-statement,expression-not-assigned


@pytest.fixture
def congress(resources):
    with Pdf.open(resources / 'congress.pdf') as pdf:
        pdfimage = pdf.pages[0].Resources.XObject['/Im0']
        yield pdfimage, pdf


def test_get_equality_stream(congress):
    image = congress[0]
    assert image.ColorSpace == image['/ColorSpace'] == image.get('/ColorSpace')
    assert image.ColorSpace == image.stream_dict.ColorSpace

    with pytest.raises(AttributeError):
        image.NoSuchKey
    with pytest.raises(KeyError):
        image['/NoSuchKey']

    assert image.get('/NoSuchKey', 42) == 42


def test_get_equality_dict(congress):
    page = congress[1].pages[0]

    assert page.MediaBox == page['/MediaBox'] == page.get('/MediaBox')

    with pytest.raises(RuntimeError):
        page.stream_dict
    with pytest.raises(AttributeError):
        page.NoSuchKey
    with pytest.raises(KeyError):
        page['/NoSuchKey']

    assert page.get('/NoSuchKey', 42) == 42



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

from typing import IO, Any, AnyStr

from lxml.etree import XMLParser as _UnsafeXMLParser
from lxml.etree import _ElementTree
from lxml.etree import parse as _parse


class _XMLParser(_UnsafeXMLParser):
    def __init__(self, *args: Any, **kwargs: Any):
        # Prevent XXE attacks
        # https://rules.sonarsource.com/python/type/Vulnerability/RSPEC-2755
        kwargs['resolve_entities'] = False
        kwargs['no_network'] = True
        super().__init__(*args, **kwargs)


def parse_xml(source: AnyStr | IO[Any], recover: bool = False) -> _ElementTree:
    """Wrap lxml's parse to provide protection against XXE attacks."""
    parser = _XMLParser(recover=recover, remove_pis=False)
    return _parse(source, parser=parser)


__all__ = ['parse_xml']



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import json

import pytest

from pikepdf import Job, JobUsageError


def test_job_from_argv(resources):
    job = Job(['pikepdf', '--check', str(resources / 'outlines.pdf')])
    job.check_configuration()
    job.message_prefix = 'foo'
    assert job.message_prefix == 'foo'
    assert not job.creates_output
    assert not job.has_warnings
    assert job.exit_code == 0
    job.run()
    assert job.exit_code == 0
    assert not job.encryption_status["encrypted"]
    assert not job.encryption_status["password_incorrect"]


def test_job_from_json(resources, outpdf):
    job_json = {}
    job_json['inputFile'] = str(resources / 'outlines.pdf')
    job_json['outputFile'] = str(outpdf)
    job = Job(json.dumps(job_json))
    job.check_configuration()
    job.run()
    assert job.exit_code == 0

    job_json = {}
    job_json['inputFile'] = str(resources / 'outlines.pdf')
    job_json['outputFile'] = str(outpdf)
    job = Job(job_json)
    job.check_configuration()
    job.run()
    assert job.exit_code == 0


def test_job_from_invalid_json():
    job_json = {}
    job_json['invalidJsonSetting'] = '123'
    with pytest.raises(RuntimeError):
        _ = Job(job_json)

    job_json2 = {}
    job_json2 = {'inputFile': []}
    with pytest.raises(JobUsageError):
        _ = Job(job_json2)


def test_schemas():
    assert isinstance(Job.LATEST_JSON, int)
    assert isinstance(Job.json_out_schema(), str)

    assert isinstance(Job.LATEST_JOB_JSON, int)
    assert isinstance(Job.job_json_schema(), str)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

import sys
from glob import glob
from itertools import chain
from os import environ
from os.path import exists, join
from platform import machine
from typing import cast

from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, build_ext
from setuptools import Extension, setup

extra_includes = []
extra_library_dirs = []
qpdf_source_tree = environ.get('QPDF_SOURCE_TREE', '')
qpdf_build_libdir = environ.get('QPDF_BUILD_LIBDIR', '')

if qpdf_source_tree:
    # Point this to qpdf source tree built with shared libraries
    extra_includes.append(join(qpdf_source_tree, 'include'))
    if not qpdf_build_libdir:
        raise Exception(
            'Please set QPDF_BUILD_LIBDIR to the directory'
            ' containing your libqpdf.so built from'
            ' $QPDF_SOURCE_TREE'
        )
    extra_library_dirs.append(join(qpdf_build_libdir))

# Here we have two different use cases. Some users will end up here because
# their package manager couldn't find a suitable binary wheel and they have to
# compile from source. Also downstream maintainers prefer sdists.
# Our priority is trying to make things work as cleanly as possible for users
# who want to do a source build. It's an imperfect test, but downstream build
# environments usually define CFLAGS to something interesting, and users who call
# "pip install pikepdf" probably don't. So we use this to check if we activate
# shims.
cflags_defined = bool(environ.get('CFLAGS', ''))
shims_enabled = not cflags_defined

# If shims are enabled, we add some known locations where QPDF and other third party
# libraries might be installed, in hopes the build will succeed if we suggest the
# obvious.
if shims_enabled and not qpdf_source_tree:
    if 'bsd' in sys.platform:
        shim_includes = ['/usr/local/include']
        shim_libdirs = []
    elif 'darwin' in sys.platform and machine() == 'arm64':
        shim_includes = ['/opt/homebrew/include', '/opt/local/include']
        shim_libdirs = ['/opt/homebrew/lib', '/opt/local/lib']
    else:
        shim_includes = []
        shim_libdirs = []
    extra_includes.extend(shim for shim in shim_includes if exists(shim))
    extra_library_dirs.extend(shim for shim in shim_libdirs if exists(shim))

# Regardless of shimming, we want to let users know when some parameter they supplied
# resulted in a non-existent path being added to the list, so they can figure out
# what went wrong.
for extra_path in chain([qpdf_source_tree], extra_includes, extra_library_dirs):
    if extra_path and not exists(extra_path):
        raise FileNotFoundError(extra_path)

# Use cast because mypy has trouble seeing Pybind11Extension is a subclass of
# Extension.
extmodule: Extension = cast(
    Extension,
    Pybind11Extension(
        'pikepdf._qpdf',
        sorted(glob('src/qpdf/*.cpp')),
        depends=sorted(glob('src/qpdf/*.h')),
        include_dirs=[
            # Path to pybind11 headers
            *extra_includes,
        ],
        define_macros=[('POINTERHOLDER_TRANSITION', '4')],
        library_dirs=[*extra_library_dirs],
        libraries=['qpdf'],
        cxx_std=17,
    ),
)

if shims_enabled:
    eca = extmodule.extra_compile_args
    if sys.platform == 'cygwin':
        # On cygwin, use gnu++17 instead of c++17
        eca[eca.index('-std=c++17')] = '-std=gnu++17'

    # Debug build
    # eca.append('-g3')

    if qpdf_source_tree:
        for lib in extra_library_dirs:
            extmodule.extra_link_args.append(f'-Wl,-rpath,{lib}')  # type: ignore

if __name__ == '__main__':
    with ParallelCompile("PIKEPDF_NUM_BUILD_JOBS"):  # optional envvar
        setup(
            ext_modules=[extmodule],
            cmdclass={"build_ext": build_ext},  # type: ignore
        )
